# coding: utf-8
# Distributed under the terms of the MIT License.

""" The workflows.castep submodule contains any workflows that rely on
CASTEP-specific implementation, for example electronic spectropscopy or phonons.

"""


__all__ = ('castep_full_phonon', 'castep_full_spectral', 'castep_elastic', 'castep_full_magres')

__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from .phonons import castep_full_phonon
from .spectral import castep_full_spectral
from .elastic import castep_elastic
from .magres import castep_full_magres



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some scraper functions for
NMR-related inputs and outputs, e.g. .magres files.

"""


from collections import defaultdict
from os import stat

import numpy as np
from matador.utils.cell_utils import cart2abc, cart2frac
from matador.utils.chem_utils import get_stoich
from matador.scrapers.utils import scraper_function, get_flines_extension_agnostic
from matador.data.constants import (
    ELECTRON_CHARGE,
    PLANCK_CONSTANT, BARN_TO_M2, EFG_AU_TO_SI
)
from matador.data.magres import ELECTRIC_QUADRUPOLE_MOMENTS


@scraper_function
def magres2dict(fname, **kwargs):
    """ Extract available information from .magres file. Assumes units of
    Angstrom and ppm for relevant quantities.
    """
    magres = defaultdict(list)
    flines, fname = get_flines_extension_agnostic(fname, "magres")
    magres['source'] = [fname]

    # grab file owner username
    try:
        from pwd import getpwuid
        magres['user'] = getpwuid(stat(fname).st_uid).pw_name
    except Exception:
        magres['user'] = 'xxx'

    magres['magres_units'] = dict()
    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ['<atoms>', '[atoms]']:
            i = 1
            while flines[line_no+i].strip().lower() not in ['</atoms>', '[/atoms]']:
                split_line = flines[line_no+i].split()
                if not split_line:
                    i += 1
                    continue
                if i > len(flines):
                    raise RuntimeError("Something went wrong in reader loop")
                if split_line[0] == 'units':
                    magres['magres_units'][split_line[1]] = split_line[2]
                elif 'lattice' in split_line:
                    lattice = split_line[1:]
                    for j in range(3):
                        magres['lattice_cart'].append([float(elem) for elem in lattice[j*3:(j+1)*3]])
                    magres['lattice_abc'] = cart2abc(magres['lattice_cart'])
                elif 'atom' in split_line:
                    atom = split_line
                    magres['atom_types'].append(atom[1])
                    magres['positions_abs'].append([float(elem) for elem in atom[-3:]])
                i += 1
            break

    if "atom_types" in magres:
        magres['num_atoms'] = len(magres['atom_types'])
        magres['positions_frac'] = cart2frac(magres['lattice_cart'], magres['positions_abs'])
        magres['stoichiometry'] = get_stoich(magres['atom_types'])

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ['<magres>', '[magres]']:
            i = 1
            while flines[line_no+i].strip().lower() not in ['</magres>', '[/magres]']:
                split_line = flines[line_no+i].split()
                if not split_line:
                    i += 1
                    continue
                if i > len(flines):
                    raise RuntimeError("Something went wrong in reader loop")
                if split_line[0] == 'units':
                    magres['magres_units'][split_line[1]] = split_line[2]
                elif 'sus' in split_line:
                    magres["susceptibility_tensor"] = np.array([float(val) for val in split_line[1:]]).reshape(3, 3)

                elif 'ms' in split_line:
                    ms = np.array([float(val) for val in split_line[3:]]).reshape(3, 3)
                    s_iso = np.trace(ms) / 3

                    # find eigenvalues of symmetric part of shielding and order them to calc anisotropy eta
                    symmetric_shielding = _symmetrise_tensor(ms)
                    s_yy, s_xx, s_zz = _get_haeberlen_eigs(symmetric_shielding)
                    s_aniso = s_zz - (s_xx + s_yy)/2.0
                    asymm = (s_yy - s_xx) / (s_zz - s_iso)

                    # convert from reduced anistropy to CSA
                    magres["magnetic_shielding_tensors"].append(ms)
                    magres["chemical_shielding_isos"].append(s_iso)
                    magres["chemical_shift_anisos"].append(s_aniso)
                    magres["chemical_shift_asymmetries"].append(asymm)

                elif "efg" in split_line:
                    efg = np.array([float(val) for val in split_line[3:]]).reshape(3, 3)
                    species = split_line[1]

                    eigs = _get_haeberlen_eigs(efg)
                    v_zz, eta = eigs[2], (eigs[0] - eigs[1]) / eigs[2]

                    # calculate C_Q in MHz
                    quadrupole_moment = ELECTRIC_QUADRUPOLE_MOMENTS.get(species, 1.0)

                    C_Q = (
                        (ELECTRON_CHARGE * v_zz * quadrupole_moment * EFG_AU_TO_SI * BARN_TO_M2)
                        / (PLANCK_CONSTANT * 1e6)
                    )

                    magres["electric_field_gradient"].append(efg)
                    magres["quadrupolar_couplings"].append(C_Q)
                    magres["quadrupolar_asymmetries"].append(eta)

                i += 1

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ['<calculation>', '[calculation]']:
            i = 1
            while flines[line_no+i].strip().lower() not in ['</calculation>', '[/calculation]']:
                if i > len(flines):
                    raise RuntimeError("Something went wrong in reader loop")
                # space important as it excludes other calc_code_x variables
                if 'calc_code ' in flines[line_no+i]:
                    magres['calculator'] = flines[line_no+i].split()[1]
                if 'calc_code_version' in flines[line_no+i]:
                    magres['calculator_version'] = flines[line_no+i].split()[1]
                i += 1

    return dict(magres), True


def _symmetrise_tensor(t):
    """ Returns the symmetrised tensor.

    Arguments:
        t: numpy array containing a NxN square tensor.

    Returns:
        np.ndarray: NxN numpy array containing the symmetric tensor.

    """
    return 0.5 * (t + t.T)


def _get_haeberlen_eigs(t: np.ndarray):
    """ Return Haeberlen convention ordered eigenvalues of the passed tensor.

    ``|s_zz - s_iso| >= |s_xx - s_iso| >= |s_yy - s_iso|``

    Arguments:
        t: numpy array containing a NxN square tensor.

    Returns:
        np.ndarray: N-d numpy array containing the ordered eigenvalues.

    """
    eig_vals, eig_vecs = np.linalg.eig(t)
    eig_vals, eig_vecs = zip(*sorted(zip(eig_vals, eig_vecs),
                                     key=lambda eig: abs(eig[0] - np.trace(t) / 3)))

    return eig_vals



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the Site class for handling
atomic sites.

"""

import numpy as np
from matador.utils.cell_utils import cart2frac, frac2cart, wrap_frac_coords
from matador.orm.orm import DataContainer


class Site(DataContainer):
    """ The Site class contains a description of an individual
    site within a 3D periodic Crystal.

    """
    # This dictionary defines the map between fields in :obj:`Crystal`
    # that correspond to arrays of site properties and between the
    # relevant keys the :obj:`Site` object
    _crystal_key_map = {
        "site_occupancy": "site_occupancy",
        "chemical_shielding_isos": "chemical_shielding_iso",
        "chemical_shift_isos": "chemical_shift_iso",
        "magnetic_shielding_tensors": "magnetic_shielding_tensor",
        "chemical_shift_anisos": "chemical_shift_aniso",
        "chemical_shift_asymmetries": "chemical_shift_asymmetry",
        "quadrupolar_couplings": "quadrupolar_coupling",
        "quadrupolar_asymmetries": "quadrupolar_asymmetry",
        "voronoi_substructure": "voronoi_substructure"
    }

    def __init__(self, species: str, position: list, lattice,
                 position_unit='fractional', **site_data):
        """ Initialise a Site object from its species, position and
        a reference to the lattice it exists in. Any other keys will be made available
        as site-level values.

        """
        if site_data.get('voronoi_substructure') is not None:
            assert self.species == site_data['voronoi_substructure'][0]
            site_data['voronoi_substructure'] = site_data['voronoi_substructure'][1]

        # DataContainer will take a copy of all data passed to it, but lets keep
        # lattice as a reference so that it can change externally
        self._lattice = lattice

        super().__init__(
            species=species,
            position=position,
            **site_data
        )

        self._data["lattice_cart"] = self._lattice

        self.set_position(position, position_unit)
        self._occupancy = None

        self.site_data = {}
        self.site_data.update(site_data)

    def __getitem__(self, key):
        """ Add extra look-up in `self.site_data` to
        :class:`DataContainer`'s `__getitem__`.

        Parameters:
            key (str): name of key or attribute to get.

        Raises:
            AttributeError: if key or attribute can't be found.

        """
        try:
            super().__getitem__(key)
        except (AttributeError, KeyError):
            pass

        try:
            return self.site_data[key]
        except KeyError:
            raise KeyError('Site has no data/site_data or implementation for requested key: "{}"'
                           .format(key))

    def __str__(self):
        site_str = '{species} {pos[0]:4.4f} {pos[1]:4.4f} {pos[2]:4.4f}'.format(species=self.species, pos=self.coords)
        for key in self.site_data:
            try:
                site_str += '\n{} = {:4.4f}'.format(key, np.asarray(self.site_data[key]))
            except (ValueError, TypeError):
                with np.printoptions(precision=2, threshold=6, edgeitems=2):
                    site_str += (
                        '\n{} = \n{}'
                        .format(
                            key,
                            "\n".join(f"  {row}" for row in np.asarray(self.site_data[key]).__str__().split("\n"))
                        )
                    )

        site_str += "\n---"

        return site_str

    def __deepcopy__(self, memo):
        from copy import deepcopy
        species, position, lattice = (deepcopy(x) for x in (self.species, self.coords, self.lattice))
        site_data = deepcopy(self.site_data)
        return Site(species, position, lattice, position_unit='fractional', **site_data)

    def set_position(self, position, units):
        if len(position) != 3 or not all(isinstance(p, (float, int)) for p in position):
            raise RuntimeError('CrystalSite position has wrong shape: {}'.format(position))
        if not hasattr(self, '_coords'):
            self._coords = dict()
        if units == 'fractional':
            self._coords['fractional'] = wrap_frac_coords(
                [float(pos) for pos in position],
                remove=False
            )
        elif units == 'cartesian':
            self._coords['fractional'] = wrap_frac_coords(
                cart2frac(self.lattice, self.coords),
                remove=False
            )
        else:
            raise RuntimeError('Unit system {} not understood, expecting `fractional`/`cartesian`'.format(units))

    @property
    def coords(self):
        return np.asarray(self._coords['fractional'])

    @property
    def coords_cartesian(self):
        return np.asarray(frac2cart(self.lattice, self.coords))

    @property
    def species(self):
        return self._data['species']

    @species.setter
    def species(self, value):
        self._data['species'] = value

    @property
    def lattice(self):
        try:
            return self._lattice.lattice_cart
        except AttributeError:
            return self._lattice

    @property
    def occupancy(self):
        if "site_occupancy" not in self._data:
            self._data["site_occupancy"] = 1.0
        return self._data["site_occupancy"]

    @property
    def coordination(self):
        if '_coordination' in self.__dict__:
            return self._coordination
        if self.voronoi_substructure is None:
            raise RuntimeError('Voronoi substructure not found.')

        coordination = {}
        eps = 0.05
        for atom, weight in self.voronoi_substructure:
            if weight >= 1 - eps:
                if atom not in coordination:
                    coordination[atom] = 1
                else:
                    coordination[atom] += 1
        self._coordination = coordination

        return coordination

    def displacement_between_sites(self, other_site):
        return self.coords_cartesian - other_site.coords_cartesian

    def distance_between_sites(self, other_site):
        return np.linalg.norm(
            self.displacement_between_sites(other_site)
        )
